# degu: an embedded HTTP server and client library
# Copyright (C) 2014 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
HTTP server.
"""

import socket
import logging
import threading
import os

from .base import bodies as default_bodies
from .base import (
    _TYPE_ERROR,
    _makefiles,
    _read_request_preamble,
    format_response_preamble,
)


log = logging.getLogger()


class UnconsumedRequestError(Exception):
    def __init__(self, body):
        self.body = body
        super().__init__(
            'previous request body not consumed: {!r}'.format(body)
        )


def build_server_sslctx(sslconfig):
    """
    Build an `ssl.SSLContext` appropriately configured for server-side use.

    For example:

    >>> sslconfig = {
    ...     'cert_file': '/my/server.cert',
    ...     'key_file': '/my/server.key',
    ...     'ca_file': '/my/client.ca',
    ... }
    >>> sslctx = build_server_sslctx(sslconfig)  #doctest: +SKIP

    """
    # Lazily import `ssl` module to be memory friendly when SSL isn't needed:
    import ssl

    if not isinstance(sslconfig, dict):
        raise TypeError(
            _TYPE_ERROR.format('sslconfig', dict, type(sslconfig), sslconfig)
        )

    # For safety and clarity, force all paths to be absolute, normalized paths:
    for key in ('cert_file', 'key_file', 'ca_file', 'ca_path'):
        if key in sslconfig:
            value = sslconfig[key]
            if value != os.path.abspath(value):
                raise ValueError(
                    'sslconfig[{!r}] is not an absulute, normalized path: {!r}'.format(
                        key, value
                    )
                )

    sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
    sslctx.set_ciphers(
        'ECDHE-RSA-AES128-GCM-SHA256:ECDHE-RSA-AES256-GCM-SHA384'
    )
    sslctx.set_ecdh_curve('secp384r1')
    sslctx.options |= ssl.OP_NO_COMPRESSION
    sslctx.options |= ssl.OP_SINGLE_ECDH_USE
    sslctx.options |= ssl.OP_CIPHER_SERVER_PREFERENCE
    sslctx.load_cert_chain(sslconfig['cert_file'], sslconfig['key_file'])
    if 'allow_unauthenticated_clients' in sslconfig:
        if sslconfig['allow_unauthenticated_clients'] is not True:
            raise ValueError(
                'True is only allowed value for allow_unauthenticated_clients'
            )
        if {'ca_file', 'ca_path'}.intersection(sslconfig):
            raise ValueError(
                'cannot include ca_file/ca_path allow_unauthenticated_clients'
            )
        return sslctx
    if not {'ca_file', 'ca_path'}.intersection(sslconfig):
        raise ValueError(
            'must include ca_file or ca_path (or allow_unauthenticated_clients)'
        )
    sslctx.verify_mode = ssl.CERT_REQUIRED
    sslctx.load_verify_locations(
        cafile=sslconfig.get('ca_file'),
        capath=sslconfig.get('ca_path'),
    )
    return sslctx


def _validate_server_sslctx(sslctx):
    # Lazily import `ssl` module to be memory friendly when SSL isn't needed:
    import ssl

    if isinstance(sslctx, dict):
        sslctx = build_server_sslctx(sslctx)

    if not isinstance(sslctx, ssl.SSLContext):
        raise TypeError('sslctx must be an ssl.SSLContext')
    if sslctx.protocol != ssl.PROTOCOL_TLSv1_2:
        raise ValueError('sslctx.protocol must be ssl.PROTOCOL_TLSv1_2')

    # We consider ssl.CERT_OPTIONAL to be a bad grey area:
    if sslctx.verify_mode == ssl.CERT_OPTIONAL:
        raise ValueError('sslctx.verify_mode cannot be ssl.CERT_OPTIONAL')
    assert sslctx.verify_mode in (ssl.CERT_REQUIRED, ssl.CERT_NONE)

    # Check the options:
    if not (sslctx.options & ssl.OP_NO_COMPRESSION):
        raise ValueError('sslctx.options must include ssl.OP_NO_COMPRESSION')
    if not (sslctx.options & ssl.OP_SINGLE_ECDH_USE):
        raise ValueError('sslctx.options must include ssl.OP_SINGLE_ECDH_USE')
    if not (sslctx.options & ssl.OP_CIPHER_SERVER_PREFERENCE):
        raise ValueError('sslctx.options must include ssl.OP_CIPHER_SERVER_PREFERENCE')

    return sslctx


def _read_request(rfile, bodies):
    (method, uri, headers) = _read_request_preamble(rfile)

    uri_parts = uri.split('?')
    if len(uri_parts) == 2:
        (path_str, query) = uri_parts
    elif len(uri_parts) == 1:
        (path_str, query) = (uri_parts[0], None)
    else:
        raise ValueError('bad request uri: {!r}'.format(uri))
    if path_str[:1] != '/' or '//' in path_str:
        raise ValueError('bad request path: {!r}'.format(path_str))
    path = ([] if path_str == '/' else path_str[1:].split('/'))

    # Only one dictionary lookup for content-length:
    content_length = headers.get('content-length')

    # Build request body:
    if content_length is not None:
        # Hack for compatibility with the CouchDB replicator, which annoyingly
        # sends a {'content-length': 0} header with all GET and HEAD requests:
        if method in {'GET', 'HEAD'} and content_length == 0:
            del headers['content-length']
        else:
            body = bodies.Body(rfile, content_length)
    elif 'transfer-encoding' in headers:
        body = bodies.ChunkedBody(rfile)
    else:
        body = None
    if body is not None and method not in {'POST', 'PUT'}:
        raise ValueError(
            'Request body with wrong method: {!r}'.format(method)
        )

    # Return the RGI request argument:
    return {
        'method': method,
        'uri': uri,
        'script': [],
        'path': path,
        'query': query,
        'headers': headers,
        'body': body,
    }


def _write_response(wfile, status, reason, headers, body):
    preamble = format_response_preamble(status, reason, headers)
    if body is None:
        total = wfile.write(preamble)
        wfile.flush()
    elif isinstance(body, (bytes, bytearray)):
        total = wfile.write(preamble + body)
        wfile.flush()
    else:
        total = wfile.write(preamble)
        total += body.write_to(wfile)          
    return total


def _handle_requests(app, sock, max_requests, session, bodies):
    (rfile, wfile) = _makefiles(sock)
    assert session['requests'] == 0
    requests = 0
    while _handle_one(app, rfile, wfile, session, bodies) is True:
        requests += 1
        session['requests'] = requests
        if requests >= max_requests:
            log.info("%r requests from %r, closing",
                requests, session['client']
            )
            break
    wfile.close()  # Will block till write buffer is flushed


def _handle_one(app, rfile, wfile, session, bodies):
    # Read the next request:
    request = _read_request(rfile, bodies)
    request_method = request['method']
    request_body = request['body']

    # Call the application:
    (status, reason, headers, body) = app(session, request, bodies)

    # Make sure application fully consumed request body:
    if request_body and not request_body.closed:
        raise UnconsumedRequestError(request_body)

    # Make sure HEAD requests are properly handled:
    if request_method == 'HEAD':
        if body is not None:
            raise TypeError(
                'response body must be None when request method is HEAD'
            )
        if 'content-length' in headers:
            if 'transfer-encoding' in headers:
                raise ValueError(
                    'cannot have both content-length and transfer-encoding headers'
                )
        elif headers.get('transfer-encoding') != 'chunked':
            raise ValueError(
                'response to HEAD request must include content-length or transfer-encoding'
            )

    # Set default content-length or transfer-encoding header as needed:
    if isinstance(body, (bytes, bytearray, bodies.Body, bodies.BodyIter)):
        length = len(body)
        if headers.setdefault('content-length', length) != length:
            raise ValueError(
                "headers['content-length'] != len(body): {!r} != {!r}".format(
                    headers['content-length'], length
                )
            )
        if 'transfer-encoding' in headers:
            raise ValueError(
                "headers['transfer-encoding'] with length-encoded body"
            )
    elif isinstance(body, (bodies.ChunkedBody, bodies.ChunkedBodyIter)):
        if headers.setdefault('transfer-encoding', 'chunked') != 'chunked':
            raise ValueError(
                "headers['transfer-encoding'] is invalid: {!r}".format(
                    headers['transfer-encoding']
                )
            )
        if 'content-length' in headers:
            raise ValueError(
                "headers['content-length'] with chunk-encoded body"
            )
    elif body is not None:
        raise TypeError(
            'body: not valid type: {!r}: {!r}'.format(type(body), body)
        )

    # Write response
    _write_response(wfile, status, reason, headers, body)

    # Possibly close the connection:
    if status >= 400 and status not in {404, 409, 412}:
        log.warning('closing connection to %r after %d %r',
            session['client'], status, reason
        )
        return False
    return True


class Server:
    _allowed_options = ('max_connections', 'max_requests', 'timeout', 'bodies')

    def __init__(self, address, app, **options):
        # address:
        if isinstance(address, tuple):  
            if len(address) == 4:
                family = socket.AF_INET6
            elif len(address) == 2:
                family = socket.AF_INET
            else:
                raise ValueError(
                    'address: must have 2 or 4 items; got {!r}'.format(address)
                )
        elif isinstance(address, str):
            if os.path.abspath(address) != address:
                raise ValueError(
                    'address: bad socket filename: {!r}'.format(address)
                )
            family = socket.AF_UNIX
        elif isinstance(address, bytes):
            family = socket.AF_UNIX
        else:
            raise TypeError(
                _TYPE_ERROR.format('address', (tuple, str, bytes), type(address), address)
            )

        # app:
        if not callable(app):
            raise TypeError('app: not callable: {!r}'.format(app))
        on_connect = getattr(app, 'on_connect', None)
        if not (on_connect is None or callable(on_connect)):
            raise TypeError('app.on_connect: not callable: {!r}'.format(app))
        self.app = app
        self.on_connect = on_connect

        # options:
        if not set(options).issubset(self.__class__._allowed_options):
            cls = self.__class__
            unsupported = sorted(set(options) - set(cls._allowed_options))
            raise TypeError(
                'unsupported {} **options: {}'.format(
                    cls.__name__, ', '.join(unsupported)
                )
            )
        self.options = options
        self.max_connections = options.get('max_connections', 25)
        self.max_requests = options.get('max_requests', 500)
        self.timeout = options.get('timeout', 30)
        self.bodies = options.get('bodies', default_bodies)
        assert isinstance(self.max_connections, int) and self.max_connections > 0
        assert isinstance(self.max_requests, int) and self.max_requests > 0 
        assert isinstance(self.timeout, (int, float)) and self.timeout > 0

        # Listen...
        self.sock = socket.socket(family, socket.SOCK_STREAM)
        self.sock.bind(address)
        self.address = self.sock.getsockname()
        self.sock.listen(5)

    def __repr__(self):
        return '{}({!r}, {!r})'.format(
            self.__class__.__name__, self.address, self.app
        )

    def serve_forever(self):
        semaphore = threading.BoundedSemaphore(self.max_connections)
        max_requests = self.max_requests
        timeout = self.timeout
        bodies = self.bodies
        listensock = self.sock
        worker = self._worker
        while True:
            (sock, address) = listensock.accept()
            # Denial of Service note: when we already have max_connections, we
            # should aggressively rate-limit the handling of new connections, so
            # that's why we use `timeout=2` rather than `blocking=False`:
            if semaphore.acquire(timeout=2) is True:
                sock.settimeout(timeout)
                thread = threading.Thread(
                    target=worker,
                    args=(semaphore, max_requests, bodies, sock, address),
                    daemon=True
                )
                thread.start()
            else:
                log.warning('Rejecting connection from %r', address)
                try:
                    sock.shutdown(socket.SHUT_RDWR)
                except OSError:
                    pass

    def _worker(self, semaphore, max_requests, bodies, sock, address):
        session = {'client': address, 'requests': 0}
        log.info('Connection from %r', address)
        try:
            self._handler(sock, max_requests, session, bodies)
        except OSError as e:
            log.info('Handled %d requests from %r: %r', 
                session.get('requests'), address, e
            )
        except:
            log.exception('Client: %r', address)
        finally:
            try:
                sock.shutdown(socket.SHUT_RDWR)
            except OSError:
                pass
            semaphore.release()

    def _handler(self, sock, max_requests, session, bodies):
        if self.on_connect is None or self.on_connect(session, sock) is True:
            _handle_requests(self.app, sock, max_requests, session, bodies)
        else:
            log.warning('rejecting connection: %r', session['client'])


class SSLServer(Server):
    def __init__(self, sslctx, address, app, **options):
        self.sslctx = _validate_server_sslctx(sslctx)
        super().__init__(address, app, **options)

    def __repr__(self):
        return '{}({!r}, {!r}, {!r})'.format(
            self.__class__.__name__, self.sslctx, self.address, self.app
        )

    def _handler(self, sock, max_requests, session, bodies):
        sock = self.sslctx.wrap_socket(sock, server_side=True)
        session.update({
            'ssl_cipher': sock.cipher(),
            'ssl_compression': sock.compression(),
        })
        super()._handler(sock, max_requests, session, bodies)




# degu: an embedded HTTP server and client library
# Copyright (C) 2014 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Pure-Python equivalent of the `degu._base` C extension.

Although `degu._basepy` is automatically imported as a fall-back when the
`degu._base` C extension isn't available, this Python implementation really
isn't meant for production use (mainly because it's much, much slower).

This is a reference implementation whose purpose is only to help enforce the
correctness of the C implementation.

Note that the Python implementation is quite different in how it decodes and
validates the HTTP preamble.  Using a lookup table is very fast in C, but is
quite slow in Python.

For *VALUES*, the Python implementation:

    1. Uses ``bytes.decode('ascii')`` to prevent bytes whose high-bit is set

    2. Uses ``str.isprintable()`` to further restrict down to the same 95 byte
       values allowed by the C ``_VALUES`` table

For *KEYS*, the Python implementation:

    1. Uses ``bytes.decode('ascii')`` to prevent bytes whose high-bit is set

    2. Uses ``str.lower()`` to case-fold the header key

    3. Uses ``re.match()`` to further restrict down to the same 63 byte values
       allowed by the C ``_KEYS`` table

Although it might seem a bit hodge-podge, this approach is much faster than
doing lookup tables in pure-Python.

However, aside from the glaring performance difference, the Python and C
implementations should always behave *exactly* the same, and we have oodles of
unit tests to back this up.

By not using lookup tables in the Python implementation, we can better verify
the correctness of the C lookup tables.  Otherwise we could have two
implementations correctly using the same incorrect tables.

``str.isprintable()`` is an especially handy gem in this respect.
"""

import re

__all__ = (
    '_MAX_LINE_SIZE',
    '_MAX_HEADER_COUNT',
    '_read_preamble',
    'EmptyPreambleError',
)

_MAX_LINE_SIZE = 4096  # Max length of line in HTTP preamble, including CRLF
_MAX_HEADER_COUNT = 20

MAX_PREAMBLE_BYTES = 65536  # 64 KiB

_RE_KEYS = re.compile('^[-0-9a-z]+$')
_RE_CONTENT_LENGTH = re.compile(b'^[0-9]+$')

_URI = frozenset(
    b'%&+-./0123456789:=?ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz~'
)

GET = 'GET'
PUT = 'PUT'
POST = 'POST'
HEAD = 'HEAD'
DELETE = 'DELETE'
OK = 'OK'


def parse_method(method):
    if isinstance(method, str):
        method = method.encode()
    if method == b'GET':
        return GET
    if method == b'PUT':
        return PUT
    if method == b'POST':
        return POST
    if method == b'HEAD':
        return HEAD
    if method == b'DELETE':
        return DELETE
    raise ValueError('bad HTTP method: {!r}'.format(method))


def _decode_value(src, message):
    """
    Used to decode and validate header values, plus the preamble first line.
    """
    text = None
    try:
        text = src.decode('ascii')
    except ValueError:
        pass
    if text is None or not text.isprintable():
        raise ValueError(message.format(src))
    return text


def _decode_key(src, message):
    """
    Used to decode, validate, and case-fold header keys.
    """
    text = None
    try:
        text = src.decode('ascii').lower()
    except ValueError:
        pass
    if text is None or not _RE_KEYS.match(text):
        raise ValueError(message.format(src))
    return text


def _decode_uri(src):
    if _URI.issuperset(src):
        return src.decode()
    raise ValueError(
        'bad uri in request line: {!r}'.format(src)
    )


class EmptyPreambleError(ConnectionError):
    pass


def _READLINE(readline, maxsize):
    """
    Matches error checking semantics of the _READLINE() macro in degu/_base.c.

    It makes sense to focus on making the pure-Python implementation a very
    correct and easy to understand reference implementation, even when at the
    expense of performance.

    So although using this _READLINE() function means a rather hefty performance
    hit for the pure-Python implementation, it helps define the correct behavior
    of the dramatically higher-performance C implementation (aka, the
    implementation you actually want to use).
    """
    assert isinstance(maxsize, int) and maxsize in (_MAX_LINE_SIZE, 2)
    line = readline(maxsize)
    if type(line) is not bytes:
        raise TypeError(
            'rfile.readline() returned {!r}, should return {!r}'.format(
                type(line), bytes
            )
        )
    if len(line) > maxsize:
        raise ValueError(
            'rfile.readline() returned {} bytes, expected at most {}'.format(
                len(line), maxsize
            )
        )
    return line


def parse_response_line(line):
    if isinstance(line, str):
        line = line.encode()

    if len(line) < 15:
        raise ValueError('response line too short: {!r}'.format(line))
    if line[0:9] != b'HTTP/1.1 ' or line[12:13] != b' ':
        raise ValueError('bad response line: {!r}'.format(line))

    # status:
    status = None
    try:
        status = int(line[9:12])
    except ValueError:
        pass
    if status is None or not (100 <= status <= 599):
        raise ValueError('bad status in response line: {!r}'.format(line))

    # reason:
    if line[13:] == b'OK':
        reason = OK
    else:
        reason = _decode_value(line[13:], 'bad reason in response line: {!r}')

    # Return (status, reason) 2-tuple:
    return (status, reason)


def parse_request_line(line):
    if isinstance(line, str):
        line = line.encode()
    if len(line) < 14:
        raise ValueError('request line too short: {!r}'.format(line))
    if line[-9:] != b' HTTP/1.1':
        raise ValueError('bad protocol in request line: {!r}'.format(line))
    line = line[:-9]
    items = line.split(b' /', 1)
    if len(items) < 2:
        raise ValueError('bad inner request line: {!r}'.format(line))
    return (
        parse_method(items[0]),
        _decode_uri( b'/' + items[1])
    )


def parse_preamble(preamble):
    (first_line, *header_lines) = preamble.split(b'\r\n')
    first_line = _decode_value(first_line, 'bad bytes in first line: {!r}')
    headers = {}
    for line in header_lines:
        (key, value) = line.split(b': ')
        key = _decode_key(key, 'bad bytes in header name: {!r}')
        value = _decode_value(value, 'bad bytes in header value: {!r}')
        if headers.setdefault(key, value) is not value:
            raise ValueError(
                'duplicate header: {!r}'.format(line)
            )
    cl = headers.get('content-length')
    if cl is not None:
        if len(cl) > 16:
            raise ValueError(
                'content-length too long: {!r}...'.format(cl[:16].encode())
            )
        try:
            value = int(cl)
        except ValueError:
            value = None
        if value is None or value < 0:
            raise ValueError(
                'bad bytes in content-length: {!r}'.format(cl)
            )
        if value > 9007199254740992:
            raise ValueError(
                'content-length value too large: {!r}'.format(value)
            )
        headers['content-length'] = value
        if 'transfer-encoding' in headers:
            raise ValueError(
                'cannot have both content-length and transfer-encoding headers'
            )
    elif 'transfer-encoding' in headers:
        if headers['transfer-encoding'] != 'chunked':
            raise ValueError(
                'bad transfer-encoding: {!r}'.format(headers['transfer-encoding'])
            )
    return (first_line, headers)


def __read_headers(readline):
    headers = {}
    for i in range(_MAX_HEADER_COUNT):
        line = _READLINE(readline, _MAX_LINE_SIZE)
        crlf = line[-2:]
        if crlf != b'\r\n':
            raise ValueError('bad header line termination: {!r}'.format(crlf))
        if line == b'\r\n':  # Stop on the first empty CRLF terminated line
            return headers
        if len(line) < 6:
            raise ValueError('header line too short: {!r}'.format(line))
        assert line[-2:] == b'\r\n'
        line = line[:-2]
        try:
            (key, value) = line.split(b': ', 1)
        except ValueError:
            key = None
            value = None
        if not (key and value):
            raise ValueError('bad header line: {!r}'.format(line))
        if key.lower() == b'content-length':
            cl = None
            if _RE_CONTENT_LENGTH.match(value):
                try:
                    cl = int(value)
                except ValueError:
                    pass
            if cl is None or cl < 0:
                raise ValueError(
                    'bad bytes in content-length: {!r}'.format(value)
                )
            if cl > 9007199254740992:
                raise ValueError(
                    'content-length value too large: {!r}'.format(cl)
                )
            key = 'content-length'
            value = cl
        elif key.lower() == b'transfer-encoding':
            if value != b'chunked':
                raise ValueError(
                    'bad transfer-encoding: {!r}'.format(value)
                )
            key = 'transfer-encoding'
            value = 'chunked'
        else:
            key = _decode_key(key, 'bad bytes in header name: {!r}')
            value = _decode_value(value, 'bad bytes in header value: {!r}')
        if headers.setdefault(key, value) is not value:
            raise ValueError(
                'duplicate header: {!r}'.format(line)
            )
    if _READLINE(readline, 2) != b'\r\n':
        raise ValueError('too many headers (> {})'.format(_MAX_HEADER_COUNT))
    return headers


def __read_preamble(rfile):
    readline = rfile.readline
    if not callable(readline):
        raise TypeError('rfile.readline is not callable')
    line = _READLINE(readline, _MAX_LINE_SIZE)
    if not line:
        raise EmptyPreambleError('HTTP preamble is empty')
    if line[-2:] != b'\r\n':
        raise ValueError('bad line termination: {!r}'.format(line[-2:]))
    if len(line) == 2:
        raise ValueError('first preamble line is empty')
    first_line = _decode_value(line[:-2], 'bad bytes in first line: {!r}')
    headers = __read_headers(readline)
    return (first_line, headers)


def _read_preamble(rfile):
    (first_line, headers) = __read_preamble(rfile)
    if 'content-length' in headers and 'transfer-encoding' in headers:
        raise ValueError(
            'cannot have both content-length and transfer-encoding headers'
        )
    return (first_line, headers)


def _read_response_preamble(rfile):
    readline = rfile.readline
    if not callable(readline):
        raise TypeError('rfile.readline is not callable')
    line = _READLINE(readline, _MAX_LINE_SIZE)
    if not line:
        raise EmptyPreambleError('HTTP preamble is empty')
    if line[-2:] != b'\r\n':
        raise ValueError('bad line termination: {!r}'.format(line[-2:]))
    if len(line) == 2:
        raise ValueError('first preamble line is empty')
    (status, reason) = parse_response_line(line[:-2])
    headers = __read_headers(readline)
    if 'content-length' in headers and 'transfer-encoding' in headers:
        raise ValueError(
            'cannot have both content-length and transfer-encoding headers'
        )
    return (status, reason, headers)


def _read_request_preamble(rfile):
    readline = rfile.readline
    if not callable(readline):
        raise TypeError('rfile.readline is not callable')
    line = _READLINE(readline, _MAX_LINE_SIZE)
    if not line:
        raise EmptyPreambleError('HTTP preamble is empty')
    if line[-2:] != b'\r\n':
        raise ValueError('bad line termination: {!r}'.format(line[-2:]))
    if len(line) == 2:
        raise ValueError('first preamble line is empty')
    (method, uri) = parse_request_line(line[:-2])
    headers = __read_headers(readline)
    if 'content-length' in headers and 'transfer-encoding' in headers:
        raise ValueError(
            'cannot have both content-length and transfer-encoding headers'
        )
    return (method, uri, headers)


class Reader:
    def __init__(self, raw):
        self.raw = raw
        self._buf = bytearray(MAX_PREAMBLE_BYTES)
        self._view = memoryview(self._buf)
        self._tell = 0
        self._size = 0

    def tell(self):
        assert isinstance(self._tell, int) and self._tell >= 0
        return self._tell

    def _consume_buffer(self, size):
        """
        Consume first *size* bytes in buffer.
        """
        assert 0 <= size <= self._size <= MAX_PREAMBLE_BYTES
        ret = self._view[:size].tobytes()
        self._tell += size
        remaining = self._size - size
        self._view[:remaining] = self._view[size:size+remaining]
        self._size = remaining
        return ret

    def _fill_buffer(self):
        assert 0 <= self._size <= len(self._buf)
        added = self.raw.readinto(self._view[self._size:])
        self._size += added
        return added

    def read_preamble(self):
        self._fill_buffer()
        index = self.buf.find(b'\r\n\r\n')
        if 0 < index < self.stop:
            pass


def format_request_preamble(method, uri, headers):
    lines = ['{} {} HTTP/1.1\r\n'.format(method, uri)]
    if headers:
        header_lines = ['{}: {}\r\n'.format(*kv) for kv in headers.items()]
        header_lines.sort()
        lines.extend(header_lines)
    lines.append('\r\n')
    return ''.join(lines).encode()


def format_response_preamble(status, reason, headers):
    lines = ['HTTP/1.1 {} {}\r\n'.format(status, reason)]
    if headers:
        header_lines = ['{}: {}\r\n'.format(*kv) for kv in headers.items()]
        header_lines.sort()
        lines.extend(header_lines)
    lines.append('\r\n')
    return ''.join(lines).encode()




# degu: an embedded HTTP server and client library
# Copyright (C) 2014 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
A few simple utility functions useful for most RGI server applications.

The module is heavily inspired by the `wsgiref.util` module in the Python3
standard library:

    https://docs.python.org/3/library/wsgiref.html
"""


def shift_path(request):
    """
    Shift path to script in an RGI *request* argument.

    For example:

    >>> request = {'script': ['foo'], 'path': ['bar', 'baz']}
    >>> shift_path(request)
    'bar'

    And you can see *request* was updated in place:

    >>> request['script']
    ['foo', 'bar']
    >>> request['path']
    ['baz']

    """
    next = request['path'].pop(0)
    request['script'].append(next)
    return next


def relative_uri(request):
    """
    Reconstruct a relative URI from an RGI *request* argument.

    For example, when there is no query:

    >>> request = {'path': ['bar', 'baz'], 'query': None}
    >>> relative_uri(request)
    '/bar/baz'

    Or when there is merely an empty query:

    >>> request = {'path': ['bar', 'baz'], 'query': ''}
    >>> relative_uri(request)
    '/bar/baz?'

    And when there is a query:

    >>> request = {'path': ['bar', 'baz'], 'query': 'stuff=junk'}
    >>> relative_uri(request)
    '/bar/baz?stuff=junk'

    Note that ``request['script']`` is ignored by this function.
    """
    uri = '/' + '/'.join(request['path'])
    query = request['query']
    if query is None:
        return uri
    return '?'.join([uri, query])


def absolute_uri(request):
    """
    Reconstruct an absolute URI from an RGI *request* argument.

    For example, when there is no query:

    >>> request = {'script': ['foo'], 'path': ['bar', 'baz'], 'query': None}
    >>> absolute_uri(request)
    '/foo/bar/baz'

    Or when there is merely an empty query:

    >>> request = {'script': ['foo'], 'path': ['bar', 'baz'], 'query': ''}
    >>> absolute_uri(request)
    '/foo/bar/baz?'

    And when there is a query:

    >>> request = {'script': ['foo'], 'path': ['bar', 'baz'], 'query': 'stuff=junk'}
    >>> absolute_uri(request)
    '/foo/bar/baz?stuff=junk'

    """
    uri = '/' + '/'.join(request['script'] + request['path'])
    query = request['query']
    if query is None:
        return uri
    return '?'.join([uri, query])




# degu: an embedded HTTP server and client library
# Copyright (C) 2014 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for the `degu.util` module`
"""

from unittest import TestCase

from degu import util


class TestFunctions(TestCase):
    def test_shift_path(self):
        # both script and path are empty:
        request = {'script': [], 'path': []}
        with self.assertRaises(IndexError):
            util.shift_path(request)
        self.assertEqual(request, {'script': [], 'path': []})

        # path is empty:
        request = {'script': ['foo'], 'path': []}
        with self.assertRaises(IndexError):
            util.shift_path(request)
        self.assertEqual(request, {'script': ['foo'], 'path': []})

        # start with populated path:
        request = {'script': [], 'path': ['foo', 'bar', 'baz']}
        self.assertEqual(util.shift_path(request), 'foo')
        self.assertEqual(request, {'script': ['foo'], 'path': ['bar', 'baz']})
        self.assertEqual(util.shift_path(request), 'bar')
        self.assertEqual(request, {'script': ['foo', 'bar'], 'path': ['baz']})
        self.assertEqual(util.shift_path(request), 'baz')
        self.assertEqual(request, {'script': ['foo', 'bar', 'baz'], 'path': []})
        with self.assertRaises(IndexError):
            util.shift_path(request)
        self.assertEqual(request, {'script': ['foo', 'bar', 'baz'], 'path': []})

    def test_relative_uri(self):
        # script, path, and query are all empty:
        request = {'script': [], 'path': [], 'query': ''}
        self.assertEqual(util.relative_uri(request), '/?')
        self.assertEqual(request, {'script': [], 'path': [], 'query': ''})

        # script and path are empty, query is None:
        request = {'script': [], 'path': [], 'query': None}
        self.assertEqual(util.relative_uri(request), '/')
        self.assertEqual(request, {'script': [], 'path': [], 'query': None})

        # only script:
        request = {'script': ['foo'], 'path': [], 'query': None}
        self.assertEqual(util.relative_uri(request), '/')
        self.assertEqual(request,
            {'script': ['foo'], 'path': [], 'query': None}
        )
        request = {'script': ['foo', ''], 'path': [], 'query': None}
        self.assertEqual(util.relative_uri(request), '/')
        self.assertEqual(request,
            {'script': ['foo', ''], 'path': [], 'query': None}
        )
        request = {'script': ['foo', 'bar'], 'path': [], 'query': None}
        self.assertEqual(util.relative_uri(request), '/')
        self.assertEqual(request,
            {'script': ['foo', 'bar'], 'path': [], 'query': None}
        )
        request = {'script': ['foo', 'bar', ''], 'path': [], 'query': None}
        self.assertEqual(util.relative_uri(request), '/')
        self.assertEqual(request,
            {'script': ['foo', 'bar', ''], 'path': [], 'query': None}
        )

        # only path:
        request = {'script': [], 'path': ['foo'], 'query': None}
        self.assertEqual(util.relative_uri(request), '/foo')
        self.assertEqual(request,
            {'script': [], 'path': ['foo'], 'query': None}
        )
        request = {'script': [], 'path': ['foo', ''], 'query': None}
        self.assertEqual(util.relative_uri(request), '/foo/')
        self.assertEqual(request,
            {'script': [], 'path': ['foo', ''], 'query': None}
        )
        request = {'script': [], 'path': ['foo', 'bar'], 'query': None}
        self.assertEqual(util.relative_uri(request), '/foo/bar')
        self.assertEqual(request,
            {'script': [], 'path': ['foo', 'bar'], 'query': None}
        )
        request = {'script': [], 'path': ['foo', 'bar', ''], 'query': None}
        self.assertEqual(util.relative_uri(request), '/foo/bar/')
        self.assertEqual(request,
            {'script': [], 'path': ['foo', 'bar', ''], 'query': None}
        )

        # only query:
        request = {'script': [], 'path': [], 'query': 'hello'}
        self.assertEqual(util.relative_uri(request), '/?hello')
        self.assertEqual(request,
            {'script': [], 'path': [], 'query': 'hello'}
        )
        request = {'script': [], 'path': [], 'query': 'stuff=junk'}
        self.assertEqual(util.relative_uri(request), '/?stuff=junk')
        self.assertEqual(request,
            {'script': [], 'path': [], 'query': 'stuff=junk'}
        )

        # All of the above:
        request = {'script': ['foo'], 'path': ['bar'], 'query': 'hello'}
        self.assertEqual(util.relative_uri(request), '/bar?hello')
        self.assertEqual(request,
            {'script': ['foo'], 'path': ['bar'], 'query': 'hello'}
        )
        request = {'script': ['foo'], 'path': ['bar', ''], 'query': 'hello'}
        self.assertEqual(util.relative_uri(request), '/bar/?hello')
        self.assertEqual(request,
            {'script': ['foo'], 'path': ['bar', ''], 'query': 'hello'}
        )
        request = {'script': ['foo'], 'path': ['bar'], 'query': 'one=two'}
        self.assertEqual(util.relative_uri(request), '/bar?one=two')
        self.assertEqual(request,
            {'script': ['foo'], 'path': ['bar'], 'query': 'one=two'}
        )
        request = {'script': ['foo'], 'path': ['bar', ''], 'query': 'one=two'}
        self.assertEqual(util.relative_uri(request), '/bar/?one=two')
        self.assertEqual(request,
            {'script': ['foo'], 'path': ['bar', ''], 'query': 'one=two'}
        )

    def test_absolute_uri(self):
        # script, path, and query are all empty:
        request = {'script': [], 'path': [], 'query': None}
        self.assertEqual(util.absolute_uri(request), '/')
        self.assertEqual(request, {'script': [], 'path': [], 'query': None})

        # only script:
        request = {'script': ['foo'], 'path': [], 'query': None}
        self.assertEqual(util.absolute_uri(request), '/foo')
        self.assertEqual(request,
            {'script': ['foo'], 'path': [], 'query': None}
        )
        request = {'script': ['foo', ''], 'path': [], 'query': None}
        self.assertEqual(util.absolute_uri(request), '/foo/')
        self.assertEqual(request,
            {'script': ['foo', ''], 'path': [], 'query': None}
        )
        request = {'script': ['foo', 'bar'], 'path': [], 'query': None}
        self.assertEqual(util.absolute_uri(request), '/foo/bar')
        self.assertEqual(request,
            {'script': ['foo', 'bar'], 'path': [], 'query': None}
        )
        request = {'script': ['foo', 'bar', ''], 'path': [], 'query': None}
        self.assertEqual(util.absolute_uri(request), '/foo/bar/')
        self.assertEqual(request,
            {'script': ['foo', 'bar', ''], 'path': [], 'query': None}
        )

        # only path:
        request = {'script': [], 'path': ['foo'], 'query': None}
        self.assertEqual(util.absolute_uri(request), '/foo')
        self.assertEqual(request,
            {'script': [], 'path': ['foo'], 'query': None}
        )
        request = {'script': [], 'path': ['foo', ''], 'query': None}
        self.assertEqual(util.absolute_uri(request), '/foo/')
        self.assertEqual(request,
            {'script': [], 'path': ['foo', ''], 'query': None}
        )
        request = {'script': [], 'path': ['foo', 'bar'], 'query': None}
        self.assertEqual(util.absolute_uri(request), '/foo/bar')
        self.assertEqual(request,
            {'script': [], 'path': ['foo', 'bar'], 'query': None}
        )
        request = {'script': [], 'path': ['foo', 'bar', ''], 'query': None}
        self.assertEqual(util.absolute_uri(request), '/foo/bar/')
        self.assertEqual(request,
            {'script': [], 'path': ['foo', 'bar', ''], 'query': None}
        )

        # only query:
        request = {'script': [], 'path': [], 'query': 'hello'}
        self.assertEqual(util.absolute_uri(request), '/?hello')
        self.assertEqual(request,
            {'script': [], 'path': [], 'query': 'hello'}
        )
        request = {'script': [], 'path': [], 'query': 'stuff=junk'}
        self.assertEqual(util.absolute_uri(request), '/?stuff=junk')
        self.assertEqual(request,
            {'script': [], 'path': [], 'query': 'stuff=junk'}
        )

        # All of the above:
        request = {'script': ['foo'], 'path': ['bar'], 'query': 'hello'}
        self.assertEqual(util.absolute_uri(request), '/foo/bar?hello')
        self.assertEqual(request,
            {'script': ['foo'], 'path': ['bar'], 'query': 'hello'}
        )
        request = {'script': ['foo'], 'path': ['bar', ''], 'query': 'hello'}
        self.assertEqual(util.absolute_uri(request), '/foo/bar/?hello')
        self.assertEqual(request,
            {'script': ['foo'], 'path': ['bar', ''], 'query': 'hello'}
        )
        request = {'script': ['foo'], 'path': ['bar'], 'query': 'one=two'}
        self.assertEqual(util.absolute_uri(request), '/foo/bar?one=two')
        self.assertEqual(request,
            {'script': ['foo'], 'path': ['bar'], 'query': 'one=two'}
        )
        request = {'script': ['foo'], 'path': ['bar', ''], 'query': 'one=two'}
        self.assertEqual(util.absolute_uri(request), '/foo/bar/?one=two')
        self.assertEqual(request,
            {'script': ['foo'], 'path': ['bar', ''], 'query': 'one=two'}
        )




# degu: an embedded HTTP server and client library
# Copyright (C) 2014 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for the `degu._basepy` module`
"""

from unittest import TestCase
import io
import os

from degu import _basepy
from degu._basepy import MAX_PREAMBLE_BYTES


class TestReader(TestCase):
    def test_init(self):
        raw = io.BytesIO(b'GET / HTTP/1.1\r\n\r\n')
        inst = _basepy.Reader(raw)
        self.assertIs(inst.raw, raw)
        self.assertIsInstance(inst._buf, bytearray)
        self.assertEqual(len(inst._buf), _basepy.MAX_PREAMBLE_BYTES)
        for i in range(_basepy.MAX_PREAMBLE_BYTES):
            self.assertEqual(inst._buf[i], 0)
        self.assertIsInstance(inst._view, memoryview)
        self.assertIs(inst._view.obj, inst._buf)
        self.assertIsInstance(inst._tell, int)
        self.assertEqual(inst._tell, 0)
        self.assertIsInstance(inst._size, int)
        self.assertEqual(inst._size, 0)

    def test_tell(self):
        raw = io.BytesIO(b'GET / HTTP/1.1\r\n\r\n')
        inst = _basepy.Reader(raw)
        self.assertEqual(inst.tell(), 0)
        inst._tell = 42
        self.assertEqual(inst.tell(), 42)

    def test_consume_buffer(self):
        inst = _basepy.Reader(None)
        data = os.urandom(42)
        inst._view[0:42] = data
        inst._size = 42

        # Consume zero bytes:
        self.assertEqual(inst._consume_buffer(0), b'')
        self.assertEqual(inst.tell(), 0)
        self.assertEqual(inst._size, 42)

        # Consume first 17 bytes:
        self.assertEqual(inst._consume_buffer(17), data[:17])
        self.assertEqual(inst.tell(), 17)
        self.assertEqual(inst._size, 25)

        # Consume first 25 bytes:
        self.assertEqual(inst._consume_buffer(25), data[17:])
        self.assertEqual(inst.tell(), 42)
        self.assertEqual(inst._size, 0)

    def test_fill_buffer(self):
        data1 = os.urandom(MAX_PREAMBLE_BYTES)
        data2 = os.urandom(34969)
        raw = io.BytesIO(data1 + data2)
        inst = _basepy.Reader(raw)

        # Test when buffer is completely empty, raw can fill buffer:
        self.assertEqual(raw.tell(), 0)
        self.assertEqual(inst._fill_buffer(), MAX_PREAMBLE_BYTES)
        self.assertEqual(raw.tell(), MAX_PREAMBLE_BYTES)
        self.assertEqual(inst._size, MAX_PREAMBLE_BYTES)
        self.assertEqual(inst._view.tobytes(), data1)

        # Test when buffer is already full:
        self.assertEqual(inst._fill_buffer(), 0)
        self.assertEqual(raw.tell(), MAX_PREAMBLE_BYTES)
        self.assertEqual(inst._size, MAX_PREAMBLE_BYTES)
        self.assertEqual(inst._view.tobytes(), data1)

        # Consume part of the buffer:
        self.assertEqual(inst._consume_buffer(34969), data1[:34969])
        self.assertEqual(raw.tell(), MAX_PREAMBLE_BYTES)
        self.assertEqual(inst._size, MAX_PREAMBLE_BYTES - 34969)

        # Test filling when raw can supply to full:
        self.assertEqual(inst._fill_buffer(), 34969)
        self.assertEqual(raw.tell(), MAX_PREAMBLE_BYTES + 34969)
        self.assertEqual(inst._size, MAX_PREAMBLE_BYTES)
        self.assertEqual(inst._view.tobytes(), data1[34969:] + data2)




#!/usr/bin/python3

import timeit

setup = """
gc.enable()

from io import BytesIO

from degu._base import (
    parse_preamble,
    parse_response_line,
    parse_request_line,
    parse_method,
)

from degu import _base, _basepy
from degu.base import bodies, _read_preamble, write_chunk
from degu.client import _write_request
from degu.server import _read_request, _write_response

line = (b'L' *  50) + b'\\r\\n'
assert line.endswith(b'\\r\\n')
assert line[-2:] == b'\\r\\n'


headers = {
    'content-type': 'application/json',
    'accept': 'application/json',
    'content-length': 12,
    'user-agent': 'Microfiber/14.04',
    'x-token': 'VVI5KPPRN5VOG9DITDLEOEIB',
    'extra': 'Super',
    'hello': 'World',
    'k': 'V',
}

fp = BytesIO()
_write_request(fp, 'POST', '/foo/bar?stuff=junk', headers, None)
request_preamble = fp.getvalue()
assert request_preamble.endswith(b'\\r\\n\\r\\n'), request_preamble
preamble = request_preamble[:-4]
del fp

data = b'D' * 1776


class wfile:
    @staticmethod
    def write(data):
        return len(data)

    @staticmethod
    def flush():
        pass

"""


def run_iter(statement, n):
    for i in range(10):
        t = timeit.Timer(statement, setup)
        yield t.timeit(n)


def run(statement, K=150):
    n = K * 1000
    # Choose fastest of 10 runs:
    elapsed = min(run_iter(statement, n))
    rate = int(n / elapsed)
    print('{:>12,}: {}'.format(rate, statement))
    return rate


print('\nSimple parsers:')
run("parse_response_line(b'HTTP/1.1 200 OK')")
run("parse_response_line(b'HTTP/1.1 404 Not Found')")
run("parse_request_line(b'GET / HTTP/1.1')")
run("parse_request_line(b'DELETE /foo/bar?stuff=junk HTTP/1.1')")
run("parse_method(b'GET')")
run("parse_method(b'PUT')")
run("parse_method(b'POST')")
run("parse_method(b'HEAD')")
run("parse_method(b'DELETE')")

print('\nFormatting request preamble:')
run("_base.format_request_preamble('GET', '/foo', {})")
run("_basepy.format_request_preamble('GET', '/foo', {})")
run("_base.format_request_preamble('PUT', '/foo', {'content-length': 17})")
run("_basepy.format_request_preamble('PUT', '/foo', {'content-length': 17})")
run("_base.format_request_preamble('PUT', '/foo', headers)")
run("_basepy.format_request_preamble('PUT', '/foo', headers)")

print('\nFormatting response preamble:')
run("_base.format_response_preamble(200, 'OK', {})")
run("_basepy.format_response_preamble(200, 'OK', {})")
run("_base.format_response_preamble(200, 'OK', {'content-length': 17})")
run("_basepy.format_response_preamble(200, 'OK', {'content-length': 17})")
run("_base.format_response_preamble(200, 'OK', headers)")
run("_basepy.format_response_preamble(200, 'OK', headers)")

print('\nFormatting and encoding:')
run("'HTTP/1.1 {} {}\\r\\n'.format(404, 'Not Found')")
run("'{} {} HTTP/1.1\\r\\n'.format('GET', '/foo/bar?stuff=junk')")
run("'{}: {}\\r\\n'.format('content-length', 1234567)")
run("'GET /foo/bar?stuff=junk HTTP/1.1\\r\\n'.encode('latin_1')")

print('\nHigh-level parsers:')
run('parse_preamble(preamble)')
run("parse_preamble(b'hello\\r\\ncontent-length: 17')")
run("parse_preamble(b'hello\\r\\ntransfer-encoding: chunked')")
run('_read_preamble(BytesIO(request_preamble))')
run('_read_request(BytesIO(request_preamble), bodies)')

print('\nHigh-level formatters:')
run("_write_response(wfile, 200, 'OK', headers, b'hello, world')")
run("_write_request(wfile, 'PUT', '/foo/bar?stuff=junk', headers, b'hello, world')")
run("write_chunk(wfile, (None, data))")
run("write_chunk(wfile, (('foo', 'bar'), data))")

print('-' * 80)




# degu: an embedded HTTP server and client library
# Copyright (C) 2014 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests the `degu.tables` module.
"""

from unittest import TestCase

from degu import tables


class TestConstants(TestCase):
    def check_allowed(self, allowed):
        self.assertIsInstance(allowed, bytes)
        self.assertEqual(len(allowed), len(set(allowed)))
        self.assertEqual(allowed, bytes(sorted(set(allowed))))
        for i in range(128):
            if not chr(i).isprintable():
                self.assertNotIn(i, allowed)
        for i in allowed:
            self.assertEqual(i & 128, 0)

    def test_KEYS(self):
        self.check_allowed(tables.KEYS)
        self.assertEqual(min(tables.KEYS), ord('-'))
        self.assertEqual(max(tables.KEYS), ord('z'))
        self.assertEqual(len(tables.KEYS), 63)

    def test_URI(self):
        self.check_allowed(tables.URI)
        self.assertEqual(min(tables.URI), ord('%'))
        self.assertEqual(max(tables.URI), ord('~'))
        self.assertEqual(len(tables.URI), 73)

    def test_VALUES(self):
        self.check_allowed(tables.VALUES)
        for i in range(128):
            if chr(i).isprintable():
                self.assertIn(i, tables.VALUES)
        self.assertEqual(min(tables.VALUES), ord(' '))
        self.assertEqual(max(tables.VALUES), ord('~'))
        self.assertEqual(len(tables.VALUES), 95)
        self.assertTrue(set(tables.VALUES).issuperset(tables.KEYS))

    def check_definition(self, definition, allowed, casefold):
        self.assertIsInstance(definition, tuple)
        self.assertEqual(len(definition), 256)
        for (index, item) in enumerate(definition):
            self.assertIsInstance(item, tuple)
            self.assertEqual(len(item), 2)
            (i, r) = item
            self.assertIsInstance(i, int)
            self.assertEqual(i, index)
            self.assertIsInstance(r, int)
            if i in allowed:
                self.assertEqual(i & 128, 0)
                if casefold:
                    self.assertEqual(r, ord(chr(i).lower()))
                else:
                    self.assertEqual(r, i)
            else:
                self.assertEqual(r, 255)
            if not (32 <= i <= 126):
                self.assertEqual(r, 255)
        self.assertEqual(definition, tuple(sorted(definition)))

    def test_KEYS_DEF(self):
        self.check_definition(tables.KEYS_DEF, tables.KEYS, True)

    def test_VALUES_DEF(self):
        self.check_definition(tables.VALUES_DEF, tables.VALUES, False)




#!/usr/bin/python3

import time
import logging
import json
import argparse
import statistics

import degu
from degu.sslhelpers import random_id
from degu.misc import TempServer
from degu.tests.helpers import TempDir
from degu.client import Client


parser = argparse.ArgumentParser()
parser.add_argument('--unix', action='store_true', default=False,
    help='Use AF_UNIX instead of AF_INET6'
)
parser.add_argument('--requests', type=int, default=10000,
    help='number of requests per connection'
)
args = parser.parse_args()


logging.basicConfig(
    level=logging.DEBUG,
    format='\t'.join([
        '%(levelname)s',
        '%(threadName)s',
        '%(message)s',
    ]),
)


agent = 'Degu/{}'.format(degu.__version__)
ping = random_id(60)
request_body = json.dumps({'ping': ping}).encode()
pong = random_id(60)
response_body = json.dumps({'pong': pong}).encode()


def ping_pong_app(session, request, bodies):
    request['body'].read()
    #assert json.loads(data.decode()) == {'ping': ping}
    return (200, 'OK', {}, response_body)


if args.unix:
    tmp = TempDir()
    address = tmp.join('my.socket')
else:
    address = degu.IPv6_LOOPBACK
server = TempServer(address, ping_pong_app, max_requests=args.requests)
client = Client(server.address)

count = args.requests
deltas = []
for i in range(10):
    conn = client.connect()
    start = time.monotonic()
    for i in range(count):
        conn.request('POST', '/', {}, request_body).body.read()
        #assert json.loads(data.decode()) == {'pong': pong}
    deltas.append(time.monotonic() - start)
    conn.close()
server.terminate()

rates = tuple(count / d for d in deltas)
fastest = '{:.2f}'.format(max(rates))
mean = '{:.2f}'.format(statistics.mean(rates))
stdev = '{:.2f}'.format(statistics.stdev(rates))
width = max(len(fastest), len(mean), len(stdev))

print('')
print('run {} of {} was fastest'.format(rates.index(max(rates)) + 1, len(rates)))
print('fastest: {} requests/second'.format(fastest.rjust(width)))
print('average: {} requests/second'.format(mean.rjust(width)))
print('  stdev: {} requests/second'.format(stdev.rjust(width)))



# degu: an embedded HTTP server and client library
# Copyright (C) 2014 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Common HTTP parser and IO abstractions used by server and client.
"""

from collections import namedtuple
try:
    from ._base import (
        _MAX_LINE_SIZE,
        _MAX_HEADER_COUNT,
        EmptyPreambleError,
        _read_response_preamble,
        _read_request_preamble,
        format_request_preamble,
        format_response_preamble,
    )
except ImportError:
    from ._basepy import (
        _MAX_LINE_SIZE,
        _MAX_HEADER_COUNT,
        EmptyPreambleError,
        _read_request_preamble,
        _read_response_preamble,
        format_request_preamble,
        format_response_preamble,
    )


__all__ = (
    '_MAX_LINE_SIZE',
    '_MAX_HEADER_COUNT',
    'EmptyPreambleError',
    '_read_response_preamble',
    '_read_request_preamble',
    'format_request_preamble',
    'format_response_preamble',
)


MAX_READ_SIZE = 16777216  # 16 MiB
MAX_CHUNK_SIZE = 16777216  # 16 MiB
STREAM_BUFFER_SIZE = 65536  # 64 KiB
IO_SIZE = 1048576  # 1 MiB

# Provide very clear TypeError messages:
_TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'


def _makefiles(sock):
    """
    Create (rfile, wfile) from a socket connection.
    """
    return (
        sock.makefile('rb', buffering=STREAM_BUFFER_SIZE),
        sock.makefile('wb', buffering=STREAM_BUFFER_SIZE)
    )


# FIXME: Add optional max_size=None keyword argument
def read_chunk(rfile):
    """
    Read a chunk from a chunk-encoded request or response body.

    See "Chunked Transfer Coding":

        http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.6.1
    """
    line = rfile.readline(_MAX_LINE_SIZE)
    if line[-2:] != b'\r\n':
        raise ValueError('bad chunk size termination: {!r}'.format(line[-2:]))
    parts = line[:-2].split(b';')
    if len(parts) > 2:
        raise ValueError('bad chunk size line: {!r}'.format(line))
    size = int(parts[0], 16)
    if not (0 <= size <= MAX_CHUNK_SIZE):
        raise ValueError(
            'need 0 <= chunk_size <= {}; got {}'.format(MAX_CHUNK_SIZE, size)
        )
    if len(parts) == 2:
        text = None
        try:
            text = parts[1].decode('ascii')  # Disallow the high-bit
        except ValueError:
            pass
        if text is None or not text.isprintable():
            raise ValueError(
                'bad bytes in chunk extension: {!r}'.format(parts[1])
            )
        (key, value) = text.split('=')
        extension = (key, value)
    else:
        extension = None
    data = rfile.read(size)
    if len(data) != size:
        raise ValueError('underflow: {} < {}'.format(len(data), size))
    crlf = rfile.read(2)
    if crlf != b'\r\n':
        raise ValueError('bad chunk data termination: {!r}'.format(crlf))
    return (extension, data)


def _encode_chunk(chunk, check_size=True):
    """
    Internal API for unit testing.
    """
    assert isinstance(chunk, tuple)
    (extension, data) = chunk
    assert extension is None or isinstance(extension, tuple)
    assert isinstance(data, bytes)
    if check_size and len(data) > MAX_CHUNK_SIZE:
        raise ValueError(
            'need len(data) <= {}; got {}'.format(MAX_CHUNK_SIZE, len(data))
        )
    if extension is None:
        size_line = '{:x}\r\n'.format(len(data))
    else:
        (key, value) = extension
        assert isinstance(key, str)
        assert isinstance(value, str)
        size_line = '{:x};{}={}\r\n'.format(len(data), key, value)
    return b''.join([size_line.encode(), data, b'\r\n'])


def write_chunk(wfile, chunk, max_size=None):
    """
    Write *chunk* to *wfile* using chunked transfer-encoding.

    Warning: the optional *max_size* keyword argument isn't yet part of the
    stable API, might go away or change in behavior.

    See "Chunked Transfer Coding":

        http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.6.1
    """
    assert isinstance(chunk, tuple)
    (extension, data) = chunk
    assert extension is None or isinstance(extension, tuple)
    assert isinstance(data, bytes)
    if max_size is None:
        max_size = MAX_CHUNK_SIZE
    if len(data) > max_size:
        raise ValueError(
            'need len(data) <= {}; got {}'.format(max_size, len(data))
        )
    if extension is None:
        size_line = '{:x}\r\n'.format(len(data))
    else:
        (key, value) = extension
        assert isinstance(key, str)
        assert isinstance(value, str)
        size_line = '{:x};{}={}\r\n'.format(len(data), key, value)
    total = wfile.write(b''.join([size_line.encode(), data, b'\r\n']))
    # Flush buffer as it could be some time before the next chunk is available:
    wfile.flush()
    return total


class Body:
    chunked = False
    __slots__ = ('rfile', 'content_length', 'io_size', 'closed', '_remaining')

    def __init__(self, rfile, content_length, io_size=None):
        if not isinstance(content_length, int):
            raise TypeError(_TYPE_ERROR.format(
                'content_length', int, type(content_length), content_length)
            )
        if content_length < 0:
            raise ValueError(
                'content_length must be >= 0, got: {!r}'.format(content_length)
            )
        if io_size is None:
            io_size = IO_SIZE
        else:
            if not isinstance(io_size, int):
                raise TypeError(
                    _TYPE_ERROR.format('io_size', int, type(io_size), io_size)
                )
            if not (4096 <= io_size <= MAX_READ_SIZE):
                raise ValueError(
                    'need 4096 <= io_size <= {}; got {}'.format(
                        MAX_READ_SIZE, io_size
                    )
                )
            if io_size & (io_size - 1):
                raise ValueError(
                    'io_size must be a power of 2; got {}'.format(io_size)
                )
        self.rfile = rfile
        self.content_length = content_length
        self.io_size = io_size
        self.closed = False
        self._remaining = content_length

    def __repr__(self):
        return '{}(<rfile>, {!r})'.format(
            self.__class__.__name__, self.content_length
        )

    def __len__(self):
        return self.content_length

    def __iter__(self):
        if self.closed:
            raise ValueError('Body.closed, already consumed')
        remaining = self._remaining
        if remaining != self.content_length:
            raise Exception('cannot mix Body.read() with Body.__iter__()')
        self._remaining = 0
        io_size = self.io_size
        read = self.rfile.read
        while remaining > 0:
            readsize = min(remaining, io_size)
            remaining -= readsize
            assert remaining >= 0
            data = read(readsize)
            if len(data) != readsize:
                self.rfile.close()
                raise ValueError(
                    'underflow: {} < {}'.format(len(data), readsize)
                )
            yield data
        self.closed = True

    # FIXME: Add 2nd, max_size=None optional keyword argument
    def read(self, size=None):
        if self.closed:
            raise ValueError('Body.closed, already consumed')
        if self._remaining <= 0:
            self.closed = True
            return b''
        if size is None:
            read = self._remaining
        else:
            if not isinstance(size, int):
                raise TypeError(
                    _TYPE_ERROR.format('size', int, type(size), size) 
                )
            if size < 0:
                raise ValueError('size must be >= 0; got {!r}'.format(size))
            read = min(self._remaining, size)
        if read > MAX_READ_SIZE:
            raise ValueError(
                'max read size exceeded: {} > {}'.format(read, MAX_READ_SIZE)
            )
        data = self.rfile.read(read)
        if len(data) != read:
            # Security note: if application-level code is being overly general
            # with their exception handling, they might continue to use a
            # connection even after an ValueError which could create a
            # request/response stream state inconsistency.  So in this
            # circumstance, we close the rfile, but we do *not* set Body.closed
            # to True (which means "fully consumed") because the body was not in
            # fact fully read. 
            self.rfile.close()
            raise ValueError(
                'underflow: {} < {}'.format(len(data), read)
            )
        self._remaining -= read
        assert self._remaining >= 0
        if size is None:
            # Entire body was read at once, so close:
            self.closed = True
        return data

    def write_to(self, wfile):
        total = sum(wfile.write(data) for data in self)
        assert total == self.content_length
        wfile.flush()
        return total


class ChunkedBody:
    chunked = True
    __slots__ = ('rfile', 'closed')

    def __init__(self, rfile):
        if not callable(rfile.read):
            raise TypeError('rfile.read is not callable: {!r}'.format(rfile))
        self.rfile = rfile
        self.closed = False

    def __repr__(self):
        return '{}(<rfile>)'.format(self.__class__.__name__)

    def __iter__(self):
        if self.closed:
            raise ValueError('ChunkedBody.closed, already consumed')
        while not self.closed:
            yield self.readchunk()

    # FIXME: Add optional max_size=None keyword argument
    def readchunk(self):
        if self.closed:
            raise ValueError('ChunkedBody.closed, already consumed')
        try:
            (extension, data) = read_chunk(self.rfile)
        except:
            self.rfile.close()
            raise
        if not data:
            self.closed = True
        return (extension, data)

    # FIXME: Add optional size=None, max_size=None keyword arguments
    def read(self, size=None):
        if self.closed:
            raise ValueError('ChunkedBody.closed, already consumed')
        buf = bytearray()
        while not self.closed:
            buf.extend(self.readchunk()[1])
            if len(buf) > MAX_READ_SIZE:
                raise ValueError(
                    'max read size exceeded: {} > {}'.format(
                        len(buf), MAX_READ_SIZE
                    )
                )
        return bytes(buf)

    def write_to(self, wfile):
        wfile.flush()  # Flush preamble before writting first chunk
        return sum(write_chunk(wfile, chunk) for chunk in self) 


class BodyIter:
    chunked = False
    __slots__ = ('source', 'content_length', 'closed', '_started')

    def __init__(self, source, content_length):
        if not isinstance(content_length, int):
            raise TypeError(_TYPE_ERROR.format(
                'content_length', int, type(content_length), content_length)
            )
        if content_length < 0:
            raise ValueError(
                'content_length must be >= 0, got: {!r}'.format(content_length)
            )
        self.source = source
        self.content_length = content_length
        self.closed = False
        self._started = False

    def __len__(self):
        return self.content_length

    def write_to(self, wfile):
        if self.closed:
            raise ValueError('BodyIter.closed, already consumed')
        if self._started:
            raise ValueError('BodyIter._started')
        self._started = True
        content_length = self.content_length
        total = 0
        for data in self.source:
            total += len(data)
            if total > content_length:
                raise ValueError(
                    'overflow: {} > {}'.format(total, content_length)
                )
            if wfile.write(data) != len(data):
                raise Exception('wfile.write() returned wrong size written')
        if total != content_length:
            raise ValueError(
                'underflow: {} < {}'.format(total, content_length)
            )
        wfile.flush()
        self.closed = True
        return total


class ChunkedBodyIter:
    chunked = True
    __slots__ = ('source', 'closed', '_started')

    def __init__(self, source):
        self.source = source
        self.closed = False
        self._started = False

    def write_to(self, wfile):
        if self.closed:
            raise ValueError('ChunkedBodyIter.closed, already consumed')
        if self._started:
            raise ValueError('ChunkedBodyIter._started')
        self._started = True
        wfile.flush()  # Flush preamble before writting first chunk
        empty = False
        total = 0
        for chunk in self.source:
            if empty:
                raise ValueError('non-empty chunk data after empty')
            total += write_chunk(wfile, chunk)
            if not chunk[1]:  # Is chunk data empty?
                empty = True
        if not empty:
            raise ValueError('final chunk data was not empty')
        self.closed = True
        return total


# Used to expose the RGI IO wrappers:
BodiesAPI = namedtuple('BodiesAPI', 'Body BodyIter ChunkedBody ChunkedBodyIter')
bodies = BodiesAPI(Body, BodyIter, ChunkedBody, ChunkedBodyIter)
